/* Terraform - (C) 1997-1999 Robert Gasch (Robert_Gasch@peoplesoft.com)
 *  - www.peoplesoft.com/peoplepages/g/robert_gasch/terraform/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "FileIO.h"
#include "HeightFieldIO.h"
#include "HeightFieldOps.h"
#include "GlobalTrace.h"
#include "GlobalSanityCheck.h"
#include "MathFunctions.h"
#include "MathGauss.h"
#include "MathRandom.h"
#include "MathTrig.h"
#include "Timer.h"
#include "tf_flexarrCOORD2.h"


#define NO_FLOW		-987654321		// marker for 'undefined'
#ifndef MIN
# define MIN(a,b)	(a<b ? a : b)
#endif
#ifndef MAX
# define MAX(a,b)	(a>b ? a : b)
#endif 


/*
 *  constructor: initialize
 */
HeightFieldOps::HeightFieldOps (HeightField *HF) 
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "+++ HeightFieldOps\n");

	SanityCheck::bailout ((!HF), "HF==NULL", "HeightFieldOps::HeightFieldOps");
	SanityCheck::bailout ((!HF->d_hf), "HF->d_hf==NULL", "HeightFieldOps::HeightFieldOps");

	d_HF = HF;
}


/*
 *  destructor: nothing to clean up
 */
HeightFieldOps::~HeightFieldOps ()
{
	GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "--- HeightFieldOps\n");
}


/*
 *  merge: provide some basic HF operators to combine 2 HFs 
 */
HeightField *HeightFieldOps::merge (char operation, HeightField *HF1, 
				HeightField *HF2, float weight1, float weight2)
{
	SanityCheck::bailout ((!HF1), "HF1==NULL", "HeightFieldOps::merge");
	SanityCheck::bailout ((!HF2), "HF2==NULL", "HeightFieldOps::merge");
	SanityCheck::bailout ((weight1 <= 0 || weight1> 1), "weight1 out of range", "HeightFieldOps::merge");
	SanityCheck::bailout ((weight2 <= 0 || weight2> 1), "weight2 out of range", "HeightFieldOps::merge");
	if (HF1->getWidth() != HF2->getWidth() ||
	    HF1->getHeight() != HF2->getHeight() )
		{
		SanityCheck::warning ((TRUE), "different size Height Fields", "HeightFieldOps::merge");
		return NULL;
		}

	int	i,
		width = HF1->getWidth(),
		height = HF1->getHeight();	
	PTYPE	*hf = new PTYPE[HF1->getSize()];

	if (operation == '+')
		{
		for (i=0; i<HF1->getSize(); i++)
			hf[i] = HF1->El(i)*weight1 + HF2->El(i)*weight2;
		}
	else
	if (operation == '-')
		{
		for (i=0; i<HF1->getSize(); i++)
			hf[i] = HF1->El(i)*weight1 - HF2->El(i)*weight2;
		}
	else
	if (operation == '*')
		{
		for (i=0; i<HF1->getSize(); i++)
			hf[i] = HF1->El(i)*weight1 * HF2->El(i)*weight2;
		}
	else
	if (operation == '/')
		{
		for (i=0; i<HF1->getSize(); i++)
			if (HF2->El(i) == 0.0)
				hf[i] = HF1->El(i)*weight1 / 0.00001*weight2;
			else
				hf[i] = HF1->El(i)*weight1 / HF2->El(i)*weight2;
		}
	else
	if (operation == 'M')		// max
		{
		for (i=0; i<HF1->getSize(); i++)
			if (HF1->El(i)*weight1 > HF2->El(i)*weight2)
				hf[i] = HF1->El(i)*weight1;
			else 
				hf[i] = HF2->El(i)*weight2;
		}
	if (operation == 'm')		// min
		{
		for (i=0; i<HF1->getSize(); i++)
			if (HF1->El(i)*weight1 < HF2->El(i)*weight2)
				hf[i] = HF1->El(i)*weight1;
			else 
				hf[i] = HF2->El(i)*weight2;
		}

	HeightField	*HF = new HeightField (hf, width, height, "preview");
	HF->setSaved (FALSE);
	return HF;
}


/*
 *  upslopeSFD: find the number of single-flow-direction 
 *	upslope cells for a given cell. Cache is used to 
 * 	to store computed SFD values.  
 */
int HeightFieldOps::upslopeSFD (int x, int y, PTYPE *cache)
{
	int	ux=x, uy=y, 		// upflow x & y coordinates
		found, count=0,
		iLow, iHigh, 
		jLow, jHigh;

	SanityCheck::bailout ((!cache), "cache==NULL", "HeightFieldOps::upslopeSFD");

	if (cache[d_HF->Pos2Off(x,y)] == NO_FLOW)
		{
		found=FALSE;
		iLow = (ux > 0 ? ux-1 : 0);
		iHigh = (ux < (d_HF->d_xsize-1) ? ux+1 : d_HF->d_xsize-1);
		jLow = (uy > 0 ? uy-1 : 0);
		jHigh = (uy < (d_HF->d_ysize-1) ? uy+1 : d_HF->d_ysize-1);
		// find the highest downflow neighbour
		for (int i=iLow; i<=iHigh; i++)
		    for (int j=jLow; j<=jHigh; j++)
			{
			if (i != x || j!= y)
				{
				if (d_HF->El(i,j) > d_HF->El(ux,uy))
					{
					found=TRUE;
					ux=i;
					uy=j;
					}
				}
			}

		// downflow
		if (found)
		    count += upslopeSFD (ux,uy,cache)+1;
		else
		    count = 0;

		cache[d_HF->Pos2Off(x,y)] = count;
		d_cMiss++;
		}
	else
		{
		count = (int)cache[d_HF->Pos2Off(x,y)];
		d_cHit++;
		}

	return (count);
}


/*
 *  upslopeMFD: find the number of multiple-flow-direction
 *	upslope cells for a given cell. countedFlag should be 
 * 	re-initialized every time flowmap() calls this 
 *	function since it's being used as an indicator 
 * 	that a cell has been counted in the current MFD 
 *	pass for this cell. Since the 
 */
float HeightFieldOps::upslopeMFD (int x, int y, PTYPE *countedFlag)
{
	int	iLow, iHigh, 
		jLow, jHigh;
	float	count=0;

	SanityCheck::bailout ((!countedFlag), "countedFlag==NULL", "HeightFieldOps::upslopeMFD");

	iLow = (x > 0 ? x-1 : 0);
	iHigh = (x < (d_HF->d_xsize-1) ? x+1 : d_HF->d_xsize-1);
	jLow = (y > 0 ? y-1 : 0);
	jHigh = (y < (d_HF->d_ysize-1) ? y+1 : d_HF->d_ysize-1);
	for (int i=iLow; i<=iHigh; i++)
	    for (int j=jLow; j<=jHigh; j++)
		{
		if (i != x || j!= y)
			{
			if ( (d_HF->El(i,j) > d_HF->El(x,y))  &&
			     (countedFlag[d_HF->Pos2Off(i,j)] == NO_FLOW) )
				{
				countedFlag[d_HF->Pos2Off(i,j)] = TRUE;
				count += upslopeMFD (i,j,countedFlag) +
					 ((i==x || j==y) ? 0.6 : 0.4);
				}
			}
		}

	return (count);
}


/*
 *  flowpap: calculate the flowmap for the height field as detailed in 
 *    "Comparison of single and mulitple flow direction algorithms for 
 *    computing topographic parameters in TOPMODEL", Wolock & McCabe Jr., 
 *    Water Resources Research, Vol. 31, No. 5, Pages 1315-1324, May 1995
 *  The flowmap is essentially   ln(a/tan(b))   where 
 *	a = A/C 
 *	C = contour length
 * 	A = (number of upslope cells + 1) * (grid cell area)
 * 	tab(b) = ( change in elevation with neighboring cell / 
 * 		   		cell center distance )
 *	ln = Napierian Log = 
 *			 	   log (n/10^7)
 *		ln (n) =     -	-------------------
 *				log (10^7/(10^7-1))
 * 
 *  Interestingly enough, the paper uses the Napierian Log, which 
 *  in my case gives huge numbers at the top of the scale and zero 
 *  at the bottom. I find that using the natural log I get a 
 *  range which closely resembles the ones shown in the paper. 
 *  Does anybody out there (who has found this) know more about this?
 *  I've probably goofed up somewhere but don't know where. 
 */
HeightField *HeightFieldOps::flowmap (bool doSFD, bool ignoreSealevel,
				float maxElvErode) 	// above this no erosion
{
	char	buf[256];
	PTYPE	ELV_RT = 0.33;
	int	i, j,
		offij,
		dfx, dfy, 			// upflow x & y coordinates
		found,	
		iLow, iHigh, 
		jLow, jHigh,
		dotint;				// dot intervals 
	double 	C=10,				// contour length XXXXXXX
		gca=C*C, 			// Grid Cell Area
		a, A,
		tanb;
	Timer	fTimer;
	float	eTime;				// elapsed time
	PTYPE	fmin=-NO_FLOW, fmax=NO_FLOW;	// min and max flowmap values
	PTYPE	*cache=NULL,		// SFD/MFD cache
		*hfFlow=NULL,			// flowmap 
		*initCache=NULL;		// cache used for memcpy source
	HeightField *HFF=NULL;		// resulting flowmap HF

	fTimer.start ();
	dotint = d_HF->d_size/50;
	d_cHit=d_cMiss=0;

	sprintf (buf, "Generating %s flowmap (%s) for %s (1 dot = %d cells) ...\n", 
		  (doSFD ? "SFD" : "MFD"), (ignoreSealevel ? "full" : "partial"), 
		  d_HF->d_name, dotint); 
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	hfFlow = new PTYPE[d_HF->d_size];	// resulting flowmap 
	cache = new PTYPE[d_HF->d_size];	// chache

	if (!doSFD)
		initCache = new PTYPE[d_HF->d_size];	// init values

	for (i=0; i<d_HF->d_size; i++)
		if (doSFD)
			cache[i] = NO_FLOW;
		else
			initCache[i] = NO_FLOW;
			
	if (!doSFD)
		memcpy (cache, initCache, sizeof(PTYPE)*d_HF->d_size);

	for (j=0; j<d_HF->d_ysize; j++)
		for (i=0; i<d_HF->d_xsize; i++)
		    if (ignoreSealevel ||
			(!ignoreSealevel && 
			  (d_HF->El(i,j) > d_HF->d_sealevel) &&
			  (d_HF->El(i,j) < maxElvErode) ) ) 
			{
			//printf ("%d, %d = ", i, j); fflush (stdout);

			if (doSFD)
				A = (upslopeSFD(i,j,cache)) * gca;
			else
				{
				memcpy (cache, initCache, 
					sizeof(PTYPE)*d_HF->d_size);
				A = (upslopeMFD(i,j,cache)) * gca;
				}
			a = A/C;

			// find lowest downslope square 
			found=FALSE;	
			dfx=i;
			dfy=j;
			iLow = (dfx > 0 ? dfx-1 : 0);
			iHigh = (dfx < (d_HF->d_xsize-1) ? dfx+1 : d_HF->d_xsize-1);
			jLow = (dfy > 0 ? dfy-1 : 0);
			jHigh = (dfy < (d_HF->d_ysize-1) ? dfy+1 : d_HF->d_ysize-1);
			for (int ii=iLow; ii<=iHigh; ii++)
			    for (int jj=jLow; jj<=jHigh; jj++)
				{
				if (ii != i || jj != j)
				    {
				    if ( d_HF->El(ii,jj) < d_HF->El(dfx,dfy) )
						{
						found = TRUE;
						dfx = ii;
						dfy = jj;
						}
				    }
				}

			// there is a downflow 
			offij = d_HF->Pos2Off(i,j);
			if (found && a)
				{
				tanb = (((d_HF->El(offij) - d_HF->El(dfx,dfy))*
					     (d_HF->d_xsize*ELV_RT)) / C);

				hfFlow[offij] = 
					//MathFunctions::napierianLog (a);
					//MathFunctions::napierianLog (1/tanb);
					//MathFunctions::napierianLog (a/tanb);
					log (a/tanb);	
					//log (a);	
				}
			// no downflow 
			else
				hfFlow[offij] = NO_FLOW;

			//printf ("Flow %d,%d, A=%f, flow=%f\n", i,j,A, hfFlow[offij]);
			if (hfFlow[offij] > fmax && hfFlow[offij]!=NO_FLOW)
				fmax = hfFlow[offij];
			if (hfFlow[offij] < fmin && hfFlow[offij]!=NO_FLOW)
				fmin = hfFlow[offij];

			if (!(offij%dotint) &&
			    GlobalTrace::trace(GlobalTrace::TRACE_VERBOSE, "."))
				;
			}

	// make 0 lowest value so that NO_FLOW==0 
	for (int i=0; i<d_HF->d_size; i++)
		{
		if (hfFlow[i] == NO_FLOW)
			hfFlow[i] = 0; 
		else
			hfFlow[i] -= fmin; 
		}
			

	fTimer.stop ();
	eTime = fTimer.getElapsedSeconds ();
	sprintf (buf, "\nDone. Elapesed time: %.2f seconds\n", eTime);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	sprintf (buf, "Flowmap max, min = %f, %f\nCache hit, miss = %ld, %ld\n", 
		fmax, fmin, d_cHit, d_cMiss);
	GlobalTrace::trace (GlobalTrace::TRACE_DEBUG, buf);

	delete [] cache;
	if (!doSFD)
		delete [] initCache;

	HFF = new HeightField (hfFlow, d_HF->d_xsize, d_HF->d_ysize);
	HFF->gatherStatistics ();
	HFF->setSaved (FALSE);

	//HeightFieldIO           *HFIO;
	//HFIO = new HeightFieldIO (buf);
	//HFIO->write (HFF);
	//delete HFIO;

	return HFF;
}


/* 
 * erode: erode the HF using the flowmap to determine the terrain delta
 */
int HeightFieldOps::erode (int 	nTurns, 	// # of erodsion turns
			   int 	maxFlowAge, 			// new flowmap every n turns
			   float maxElvErode, 			// above this no erosion
			   char *fName,					// if fName = save every maxFlowAge turns
			   bool trimLocalPeaks,			// trim if (flowmap==0)
			   bool doSFD, 					// flowmap param
			   bool ignoreSealevel, 		// flowmap param
			   HeightFieldDraw *HFD)
{
	char		buf[256];
	HeightField 	*HFF;			// flowmap HF
	int		nFlowmap=0,				// number of flowmaps so far
			flowAge=0,				// age of flowmap
			dotint;
	double		errate = 2e-4;		// max erosion per turn

	dotint = int(maxFlowAge/50.0);

	SanityCheck::bailout ((!HFD), "HFD==NULL", "HeightFieldOps::erode");
	// fName should be formatted as a sprintf format string and contain %d
	if (fName)
		{
		SanityCheck::bailout ((!strstr(fName,"%d")), "bad string format", "HeightFieldOps::erode");
		SanityCheck::bailout ((!strchr(fName, '.')), "bad string format", "HeightFieldOps::erode");
		}

	HFF = flowmap (doSFD, ignoreSealevel);
	for (int i=0; i<nTurns; i++, flowAge++)
		{
		if (!(i%dotint) &&
		    GlobalTrace::trace(GlobalTrace::TRACE_FLOW, "."))
			;

		d_HF->d_hf[0] -= errate*HFF->d_hf[0];
		for (int n=1; n<d_HF->d_size; n++)
			{
			if (HFF->d_hf[n] > 0)
				d_HF->d_hf[n] -= errate*HFF->d_hf[n];
			else
				{
				// average flowmap vale with neighbours
				if (trimLocalPeaks)
					d_HF->d_hf[n] -= errate * 
						( HFF->El(n-1) + 
						  HFF->El(n+1) + 
						  HFF->El(n-d_HF->d_xsize) + 
						  HFF->El(n+d_HF->d_xsize) )/4;
				else
					d_HF->d_hf[n] -= errate*HFF->El(n);
				}

			if (d_HF->d_hf[n] < 0)
				d_HF->d_hf[n] = 0;
			}

		if ((maxFlowAge>0) && (flowAge == maxFlowAge))
			{
			GlobalTrace::trace (GlobalTrace::TRACE_FLOW, "\n");
			//roughSmooth (false, true, 1.0);
			d_HF->gatherStatistics ();
			if (HFD)
				HFD->draw ();

			if (fName)
				{
				sprintf (buf, fName, nFlowmap);
				HeightFieldIO           *HFIO;

				HFIO = new HeightFieldIO (buf);
				HFIO->write (d_HF);
				delete HFIO;  
				}

			delete HFF;
			HFF = flowmap (doSFD, ignoreSealevel);
			flowAge = 0;
			nFlowmap++;

			sprintf (buf, "\nEroding %d/%d\n", i, nTurns); 
			GlobalTrace::trace (GlobalTrace::TRACE_FLOW, buf);
			}
		}

	if (HFF)
		delete HFF;

	return 0;
}



/*
 * hfDouble()  double resolution using modified-midpoint-displacement
 *	       algorithm; different versions for tiling and non-tiling
 *	       matrices. <f> is local fractal scaling, <f2> is overall
 *	       fractal scaling.
 * 	       This routine is adapted from John Beals's HF-Lab.
 */
int HeightFieldOps::hfDouble (double f, double f2) 	
{
	char		buf[80];
	int 		x, y;
	int 		xsize=0, ysize=0;
	int 		xsize1, ysize1;   	// old1 HF 
	int 		xsize2, ysize2;   	// old2 HF
  	PTYPE 		*hf1=d_HF->d_hf, 
			*hf2;
  	int 		tile=FALSE;
  	D 		sfac;
  	D 		a, b, c, d;
	MathGauss	*mGauss = new MathGauss (time(NULL));

	xsize1 = d_HF->d_xsize;
	ysize1 = d_HF->d_ysize;

	if (!tile) 			// non-tiling: don't wrap at edges 
		{  
		xsize2--;
		ysize2--;
		}

	d_HF->d_xsize = xsize2 = xsize1*2;
	d_HF->d_ysize = ysize2 = ysize1*2;
	d_HF->d_size = xsize2*ysize2;

	sprintf (buf, "doubling to %dx%d ... ", xsize2, ysize2);  
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	hf2 = new PTYPE[d_HF->d_size];
	d_HF->d_hf = hf2;

	f2 /= sqrt((D)d_HF->d_size);  /* normalize to matrix size */

	for (y=0;y<ysize1;y++) 
     		for (x=0;x<xsize1;x++)
      			hf2[d_HF->Pos2Off(x*2,y*2)] = hf1[xsize1*y+x];

	if (!tile) 			// fill in edges 
		{
		for (y=1;y<ysize2;y+=2) 
			{
			hf2[d_HF->Pos2Off(0,y)] = hf2[d_HF->Pos2Off(0,y-1)];
			hf2[d_HF->Pos2Off(xsize2-1,y)] = hf2[d_HF->Pos2Off(xsize2-1,y-1)];
			}
		for (x=1;x<xsize2;x+=2) 
			{
			hf2[d_HF->Pos2Off(x,0)] = hf2[d_HF->Pos2Off(x-1,0)];
			hf2[d_HF->Pos2Off(x,ysize2-1)] = hf2[d_HF->Pos2Off(x-1,ysize2-1)];
			}
		}

	xsize = xsize2;
	ysize = ysize2;
	for (y=1; y<ysize2; y+=2) 	// tiling and non-tiling interpolation
	    {
	    for (x=1; x<xsize2; x+=2) 
		{
		if (tile) 
			{
			a = d_HF->Elmod(x+1,y+1); b = d_HF->Elmod(x-1,y+1);
			c = d_HF->Elmod(x+1,y-1); d = d_HF->Elmod(x-1,y-1);
			} 
		else 
			{
			a = d_HF->Elclip(x+1,y+1); b = d_HF->Elclip(x-1,y+1);
			c = d_HF->Elclip(x+1,y-1); d = d_HF->Elclip(x-1,y-1);
			}
	        sfac = f2+f*(MAX(MAX(a,b),MAX(c,d)) - MIN(MIN(a,b),MIN(c,d)));
		d_HF->d_hf[d_HF->Pos2Off(x,y)] = (a+b+c+d)/4.0+sfac*(mGauss->rnd()-0.5);
		}
	    }

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "cent1 ...");

	for (y=0; y<ysize2; y+=2) 
	    {      
	    for (x=1; x<xsize2; x+=2) 
		{
		if (tile) 
			{
			a = d_HF->Elmod(x+1,y); b = d_HF->Elmod(x-1,y);
			c = d_HF->Elmod(x,y+1); d = d_HF->Elmod(x,y-1);
			} 
		else 
			{
			a = d_HF->Elclip(x+1,y); b = d_HF->Elclip(x-1,y);
			c = d_HF->Elclip(x,y+1); d = d_HF->Elclip(x,y-1);
			}
		sfac = f2+f*(MAX(MAX(a,b),MAX(c,d)) - MIN(MIN(a,b),MIN(c,d)));
		d_HF->d_hf[d_HF->Pos2Off(x,y)] = (a+b+c+d)/4.0+sfac*(mGauss->rnd()-0.5);
		}
	    }

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "edge1 ...");

	for (y=1; y<ysize2; y+=2) 
	    {      
	    for (x=0; x<xsize2; x+=2) 
		{
		if (tile) 
			{
			a = d_HF->Elmod(x+1,y); b = d_HF->Elmod(x-1,y);
			c = d_HF->Elmod(x,y+1); d = d_HF->Elmod(x,y-1);
			} 
		else 
			{
			a = d_HF->Elclip(x+1,y); b = d_HF->Elclip(x-1,y);
			c = d_HF->Elclip(x,y+1); d = d_HF->Elclip(x,y-1);
			}
		sfac = f2+f*(MAX(MAX(a,b),MAX(c,d)) - MIN(MIN(a,b),MIN(c,d)));
		d_HF->d_hf[d_HF->Pos2Off(x,y)] = (a+b+c+d)/4.0+sfac*(mGauss->rnd()-0.5);
		}
	    }

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "edge2 ...");

	for (y=0;y< ysize2; y+=2) 
	    {      /* backtrack and adjust original data */
	    for (x=0; x<xsize2; x+=2) 
		{
		if (tile) 
			{
			a = d_HF->Elmod(x+1,y+1); b = d_HF->Elmod(x-1,y+1);
			c = d_HF->Elmod(x+1,y-1); d = d_HF->Elmod(x-1,y-1);
			} 
		else 
			{
			a = d_HF->Elclip(x+1,y+1); b = d_HF->Elclip(x-1,y+1);
			c = d_HF->Elclip(x+1,y-1); d = d_HF->Elclip(x-1,y-1);
			}
		sfac = 0.5*(f2+f*(MAX(MAX(a,b),MAX(c,d)) - MIN(MIN(a,b),MIN(c,d))));
		d_HF->d_hf[d_HF->Pos2Off(x,y)] = (a+b+c+d)/4.0+sfac*(mGauss->rnd()-0.5);
		}
	    }

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "done\n");

	delete [] hf1;
	delete mGauss;

	d_HF->gatherStatistics ();
	d_HF->setSaved (FALSE);
	return (0);
}  



/*
 * hfHalf ()   half resolution by averaging 4 points into 1
 * 	       This routine is adapted from John Beals's HF-Lab.
 */
int HeightFieldOps::hfHalf() 
{
	int 	x,y,xx,yy;
	int 	xsize1, ysize1;   /* old1 HF */
	int 	xsize2, ysize2;   /* old2 HF */
  	PTYPE 	*hf1=d_HF->d_hf, *hf2;
	char	buf[80];

#define P2O(x,y)	y*xsize1+x

	if (d_HF->d_xsize*d_HF->d_ysize < 25)
		return 1;

	xsize1 = d_HF->d_xsize;
	ysize1 = d_HF->d_ysize;
	d_HF->d_xsize = xsize2 = xsize1/2;
	d_HF->d_ysize = ysize2 = ysize1/2;
	d_HF->d_size = xsize2*ysize2;

	sprintf(buf, "Diminishing to %dx%d ... ", xsize2,ysize2); 
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	hf2 = new PTYPE[d_HF->d_size];
	d_HF->d_hf = hf2;

	for (y=0; y<ysize2; y++) 
	    {      /* copy over initial values */
	    yy = y * 2;
	    for (x=0; x<xsize2; x++) 
		{
		xx = x * 2;
		d_HF->d_hf[d_HF->Pos2Off(x,y)] = 
			( hf1[P2O(xx,yy)] + hf1[P2O(xx+1,yy)] +
			  hf1[P2O(xx,yy+1)] + hf1[P2O(xx+1,yy+1)] ) /4.0;
		}
	    }

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "done\n");

	delete [] hf1;
	d_HF->gatherStatistics ();
	d_HF->setSaved (FALSE);
	return (0);
}  


/*
 * hfResize()   half resolution by averaging 4 points into 1
 * 	       This routine is adapted from John Beals's HF-Lab.
 */
HeightField *HeightFieldOps::hfResize (int newx, int newy)
{
	float 		incx = ((float)d_HF->d_xsize)/((float)newx),
			incy = d_HF->d_ysize/newy;
	int		x, 
			y,
			offSrc,
			offDst;
	HeightField	*HF=NULL;

	if (d_HF->d_xsize<=newx || d_HF->d_ysize<newy)
		return NULL;

	PTYPE	*hf = new PTYPE[newx*newy];

	for (y=0; y<newy; y++)
		for (x=0; x<newx; x++)
			{
			offSrc = d_HF->Pos2Off ((int)(x*incx),(int)(y*incy));
			offDst = (int)(y*newx+x);
			hf[offDst] = d_HF->El (offSrc);
			//printf ("HFResize: %d  ->  %f\n", offDst, hf[offDst]);
			}

	HF = new HeightField (hf, newx, newy, d_HF->getName());
	HF->setSaved (FALSE);

	return HF;
}



/*
 *  placePOVObject: place a PovRay object on the landscape
 * 	We use the following method:
 *	- calculate a basic square size dependant upon object placement density and HF scale size in POV units.
 *	- calculate the size of an inner square proportional to placeVar 
 *	- place an object of randomized size at the correct offset for the inner square
 */
int HeightFieldOps::placePOVObject (
			char	*outfile,	// output filename
			char 	*object, 	// object name
			int	hfScaleX,	// height field x scaling factor
			int	hfScaleY,	// height field y scaling factor
			int	HFScaleZ,	// height field y scaling factor
			PTYPE 	floor, 		// low elevation for object placement
			PTYPE	ceil, 		// high elevation for object placement
			float	density,	// density factor for object placement
			float	placeVar,	// placement variance ratio 
			int	objScaleX, 	// object x scaling factor
			int	objScaleY, 	// object y scaling factor
			int	objScaleZ, 	// object y scaling factor
			PTYPE	objScaleXvar, 	// object x scaling factor variance
			PTYPE	objScaleYvar, 	// object z scaling factor variance
			PTYPE	objScaleZvar, 	// object z scaling factor variance
			int	seed)		// random number seed
{
	FILE	*fout, *ft;
	float	e,				// elevation
		xOff, yOff,			// offset for inner placement square
		nGridX, nGridY,			// numbers of x and y size of a grid
		cellSizeX, cellSizeY,			// x and y size of a grid
		ixpov, izpov, 			// x and y increment POVray scale
		ixhf, iyhf, 			// x and y increment d_HF subscript scale
		hsx, hsy, hsz, lsx, lsy, lsz;	// high and low scaling factors
	int	nobs=0;
	char	buf[80];
    	MathRandom	*rnd = NULL;

	//printf ("\n%s, %s, hfs[xyz]=%d, %d, %d\n", outfile, object, hfScaleX, hfScaleY, HFScaleZ);
	//printf ("floor=%f, ceil=%f, dens=%f, placeVar=%f,\n", floor, ceil, density, placeVar);
	//printf ("os[xyz]=%d,%d,%d ", objScaleX, objScaleY, objScaleZ);
	//printf ("os[xyz]var=%f, %f, %f, seed=%d\n\n", objScaleXvar, objScaleYvar, objScaleZvar, seed);

	// Sanity Checks
	SanityCheck::bailout ((!outfile), "outfile==NULL", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((!strlen(outfile)), "strlen(outfile)==0", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((!object), "object==NULL", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((!strlen(object)), "strlen(object)==0", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((floor < d_HF->d_min || floor > d_HF->d_max), 
			"floor out of range", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((ceil < d_HF->d_min || ceil > d_HF->d_max), 
			"ceil out of range", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((density <= 0), "density<=0", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((placeVar < 0 || placeVar > 1), "placeVar out of range", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((objScaleXvar < 0 || objScaleXvar > 1), "objScaleXvar out of range", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((objScaleYvar < 0 || objScaleYvar > 1), "objScaleYvar out of range", "HeightFieldOps::PlacePOVObject");
	SanityCheck::bailout ((objScaleZvar < 0 || objScaleZvar > 1), "objScaleZvar out of range", "HeightFieldOps::PlacePOVObject");

	if (!(fout=fopen (outfile, "w"))) 		// open output file
		{
		cerr << "HeightFieldOps::PlacePOVObject(): can't open file " << outfile << "\n";
		return (-1);
		}

    	rnd = new MathRandom (seed);

	sprintf (buf, "Placing PovObjects on %s ... ", d_HF->d_name);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	// find percentage scaling ranges (0 .. 2)
	hsx=1+objScaleXvar; hsy=1+objScaleYvar; hsz=1+objScaleZvar;
	lsx=1-objScaleXvar; lsy=1-objScaleYvar; lsz=1-objScaleZvar;

	// calculate number of cells 
	nGridX=(hfScaleX/objScaleX)*density;		// high density == more cells
	nGridY=(HFScaleZ/objScaleZ)*density;		// povray dim z == program dim y

	// calculate cell size 
	cellSizeX=(hfScaleX/nGridX);
	cellSizeY=(HFScaleZ/nGridY);

	// calculate max offset from center
	xOff=(cellSizeX-(cellSizeX-(cellSizeX*placeVar)));
	yOff=(cellSizeY-(cellSizeY-(cellSizeY*placeVar)));

	printf ("%f, %f, %f, %f\n", nGridX, nGridY, cellSizeX, cellSizeY); 
	printf ("%f, %f\n", xOff, yOff);

	ft=stdout;
	for (int c=0; c<2; c++)
	   {
	   if ((c==0 && GlobalTrace::isSet (GlobalTrace::TRACE_DEBUG)) || c==1)
		{
		if (c==1)
			fprintf (ft, "// File created by terraform ...\n");
		fprintf (ft, "// HeightFieldOps::PlacePOVObject Settings\n");
		fprintf (ft, "// Output Filename: 		%s\n", outfile);
		fprintf (ft, "// Object name:	 		%s\n", object);
		fprintf (ft, "// Height Field Scales (x,y,z):	%d, %d, %d\n", hfScaleX, hfScaleY, HFScaleZ);
		fprintf (ft, "// Floor, Ceil, Density:	%f, %f, %f\n", floor, ceil, density);
		fprintf (ft, "// x scale & variance:		%d, %f\n", objScaleX, objScaleXvar);
		fprintf (ft, "// y scale & variance:		%d, %f\n", objScaleY, objScaleYvar);
		fprintf (ft, "// z scale & variance:		%d, %f\n", objScaleZ, objScaleZvar);
		fprintf (ft, "// seed:			%d\n", seed);
		fprintf (ft, "// hsx, hsy, hsz:			%f, %f, %f\n", hsx, hsy, hsz);
		fprintf (ft, "// lsx, lsy, lsz:			%f, %f, %f\n", lsx, lsy, lsz);
		fprintf (ft, "// nGridX, nGridY:		%f, %f\n", nGridX, nGridY);
		fprintf (ft, "// xOff, yOff:			%f, %f\n\n", xOff, yOff);
		}
	   ft=fout;
	   }

	ixpov=cellSizeX;			// increment on POVray scale
	izpov=cellSizeY;			// povray dim z == array dim y
	ixhf=d_HF->d_xsize/nGridX;		// increment on d_HF subscript scale
	iyhf=d_HF->d_ysize/nGridY;		// increment on d_HF subscript scale

	
	printf ("xOff=%f, yOff=%f\nnGridX=%f, nGridY=%f\n", xOff, yOff, nGridX, nGridY);
	printf ("ixpov=%f, izpov=%f, ixhf=%f, iyhf=%f\n", ixpov, izpov, ixhf, iyhf);

	// loop through the squares on Pov and Hf offsets
	for (float jp=0, jh=0; jp<HFScaleZ; jp+=izpov, jh+=iyhf)
		for (float ip=0, ih=0; ip<hfScaleX; ip+=ixpov, ih+=ixhf)
			{
			// variables holding placement info for current object 
			float	aoffx, aoffz,	// actual object placement offset in grid (from grid center)
				aobjScaleX=0, 	// actual object scales 
				aobjScaleY=0, 
				aobjScaleZ=0,	
				aopx=0, 	// actual object position 
				aopy=0, 
				aopz=0;	

			//printf ("X:: %f, %f, %f, %f, %d", ip, jp, ih, jh, off);fflush(stdout);

			// reverse Y/Z coordinate system for POV 
			e=d_HF->d_hf[d_HF->Pos2Off(ih,d_HF->d_ysize-jh)];	// reverse drawing direction from terraform to povray

			if (e >= floor && e <= ceil)
				{
				nobs++;

#ifdef TEST_CODE
				// this will place objects with no variance
				// was useful for testing 
				aopx = ip;
				aopz = jp;
				aopy = e*hfScaleY;
				fprintf (fout, "object { %s  scale < %f, %f, %f > translate < %f, %f, %f > }\n",
					object, 1.0, 1.0, 1.0, aopx, aopy, aopz);
#else

				// get an offset for the inner placement square
				aoffx = xOff*rnd->rnd1();
				aoffz = yOff*rnd->rnd1();

				// calculate the object position in the scene
				if (rnd->rnd1() > 0.5)
					aopx = ip+ixpov/2+aoffx;
				else
					aopx = ip+ixpov/2-aoffx;

				if (rnd->rnd1() > 0.5)
					aopz = jp+izpov/2+aoffz;
				else
					aopz = jp+izpov/2-aoffz;

				aopy = e*hfScaleY;
	
				// calculate the object scaling factors
				aobjScaleX = ((hsx-lsx)*rnd->rnd1()+lsx);
				aobjScaleY = ((hsy-lsy)*rnd->rnd1()+lsy);
				aobjScaleZ = ((hsz-lsz)*rnd->rnd1()+lsz);

				fprintf (fout, "object { %s  scale < %f, %f, %f > translate < %f, %f, %f > }\n",
					object, objScaleX*aobjScaleX, 
					objScaleY*aobjScaleY, objScaleZ*aobjScaleZ, 
					aopx, aopy, aopz);
#endif
				}
			}
	fclose (fout); 			// close output file

	sprintf (buf, "%d %ss placed\n", nobs, object);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	delete rnd;
	return (0);
}


/*
 *  roughSmooth: either roughen or smooth the height field. 
 */
int HeightFieldOps::roughSmooth (bool roughen, bool bigGrid, float factor)
{
	SanityCheck::bailout ((factor<0), "factor<0", "HeightFieldOps::smooth");
	SanityCheck::bailout ((factor>1), "factor>1", "HeightFieldOps::smooth");

	PTYPE	o, t;
	int	limx=d_HF->d_xsize-1, limy=d_HF->d_ysize-1;

	for (int x=1; x<limx; x++)
		for (int y=1; y<limy; y++)
			{
			o = d_HF->El (x, y);
			if (bigGrid == 0)
				{
				t = d_HF->El(x-1,y) + d_HF->El(x+1,y) +
				    d_HF->El(x,y-1) + d_HF->El(x,y+1);
				t /= 4;
				}
			else
				{
				t = d_HF->El(x-1,y) + d_HF->El(x+1,y) +
				    d_HF->El(x,y-1) + d_HF->El(x,y+1) + 
				    d_HF->El(x-1,y-1) + d_HF->El(x-1,y+1) +
				    d_HF->El(x+1,y-1) + d_HF->El(x+1,y+1);
				t /= 8;
				}
			if (roughen)
				d_HF->d_hf[d_HF->Pos2Off(x,y)] = o-factor*(t-o);
			else
				d_HF->d_hf[d_HF->Pos2Off(x,y)] = o+factor*(t-o);
			}

	d_HF->gatherStatistics ();
	d_HF->setSaved (FALSE);
	return (0);
}


/*
 *  transform: transform the elevation by some power factor
 */
int HeightFieldOps::transform (float factor, int adjust_sea_level)
{
	SanityCheck::bailout ((!d_HF->d_hf), "d_HF->d_hf==NULL", "HeightFieldOps::transform");

	char	buf[80];

	sprintf (buf, "Transforming with power %f ...\n", factor);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	for (int i=0; i<d_HF->d_size; i++)
		d_HF->d_hf[i]=pow (d_HF->d_hf[i], factor);
	d_HF->gatherStatistics ();
	if (adjust_sea_level)
		d_HF->setSealevel (pow (d_HF->d_sealevel, factor));;
	d_HF->setSaved (FALSE);
	return (0);
}


/*
 *  terrace: make terraces into the terrain
 */
int HeightFieldOps::terrace (int nLevels, float factor, bool adjustSealevel)
{
	SanityCheck::bailout ((!d_HF->d_hf), "d_HF->d_hf==NULL", "HeightFieldOps::transform");

	// increase nLevels by one so that the highest elevation (1.0) gets 
	// the last level (which is 1 point high).
	nLevels++;
	factor=1-factor;		// convert tightness scale 

	char	buf[80];
	int	level;
	float	range = d_HF->d_max - d_HF->d_min,
		lRange = range/nLevels,
		base,
		diff;


	sprintf (buf, "Terracing %d levels with tightness %f ...\n", nLevels, factor);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	for (int i=0; i<d_HF->d_size; i++)
		{
		level = (int)(d_HF->El(i)/lRange);
		base = level*lRange;
		diff = d_HF->El(i) - base;
		d_HF->d_hf[i] = base + diff*factor;
		}

	d_HF->gatherStatistics ();
	d_HF->setSaved (FALSE);
	return (0);
}


/*
 *  transform: transform the elevation by some power factor
 */
int HeightFieldOps::gaussianHill (float cx, float cy, float radius, 
				float radfac, float hscale, 
				float smooth, float deltascale)
{
	int 	xsize = d_HF->getWidth(), 
			ysize = d_HF->getHeight();
	float	d, 								// distance
			sf, 							// smoothing factor
			fxmin, fxmax, fymin, fymax, 	// min, max working floats
			sdist, sdistr;					// smoothing distances 
 	int 	xmin, xmax, ymin, ymax, xcent, ycent;
 	int 	x, y;
	D 		xa,ya,tmp,rad;
	char	buf[80];

	SanityCheck::bailout ((cx<0), "cx<0", "HeightFieldOps::gaussianHill");
	SanityCheck::bailout ((cx>1), "cx>1", "HeightFieldOps::gaussianHill");
	SanityCheck::bailout ((cy<0), "cy<0", "HeightFieldOps::gaussianHill");
	SanityCheck::bailout ((cy>1), "cy>1", "HeightFieldOps::gaussianHill");
	SanityCheck::bailout ((smooth<0), "smooth<0", "HeightFieldOps::gaussianHill");
	SanityCheck::bailout ((smooth>1), "smooth>0.5", "HeightFieldOps::gaussianHill");
	SanityCheck::bailout ((deltascale<0), "deltascale<0", "HeightFieldOps::gaussianHill");
	SanityCheck::bailout ((deltascale>1), "deltascale>1", "HeightFieldOps::gaussianHill");

	// calculate max and min coordinates on 0 .. 1 scale
	fxmin = cx - radius;
	fxmax = cx + radius;
	fymin = cy - radius;
	fymax = cy + radius;
	if (fxmin < 0) fxmin = 0;
	if (fxmax > 1) fxmax = 1;
	if (fymin < 0) fymin = 0;
	if (fymax > 1) fymax = 1;

	// translate everything to subscript scale (0 .. n)
	xmin = (int)(fxmin*xsize);	
	xmax = (int)(fxmax*xsize);
	ymin = (int)(fymin*ysize);
	ymax = (int)(fymax*ysize);
	xcent = (int)(cx*xsize);
	ycent = (int)(cy*xsize);
	radius *= ((xsize+ysize)/2);
	sdist = radius*smooth;
	sdistr = radius-sdist;

	sprintf(buf, "Gaussian Hill: %f, %f: radfac=%f, hscale=%f\n", 
		cx, cy, radfac, hscale);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);
	sprintf(buf, "Gaussian Hill: xrange=%d,%d, yrange=: %d,%d\n", 
		xmin, xmax, ymin, ymax);
	GlobalTrace::trace (GlobalTrace::TRACE_DEBUG, buf);

 	radfac = radfac * radfac;
	for (y=ymin; y<ymax; y++) 
		{
		ya = (((D)y+0.5)/ysize) - cy;
		ya = ya*ya;

		for (x=xmin; x<xmax; x++) 
			{
			d = MathTrig::distance (xcent, ycent, x, y);
			if (d <= radius)
				{
    				tmp = d_HF->El(x,y); 
				xa = (((D)x+0.5)/xsize) - cx;
				xa = xa * xa;
				rad = xa+ya;

				// see if we should be smoothing
				if (d > sdistr)
					{
					d -= sdistr;
					sf = 1-d/sdist;
					tmp += deltascale * sf * 
						hscale * exp(-rad/radfac);
					}
				else
					tmp += deltascale * hscale * 
						exp(-rad/radfac);
				d_HF->d_hf[d_HF->Pos2Off(x,y)] = tmp;
    				//printf ("%d  -->  %f\n", d_HF->Pos2Off(x,y), d_HF->El(x,y)); 
				}
			} 
		} 

	d_HF->gatherStatistics ();
	d_HF->setSaved (FALSE);
	return (0);
}


/*
 *  rotate90: counterclockwise rotate in 90, 180 or 270 degrees
 */
int HeightFieldOps::rotate90 (int degrees)
{
	SanityCheck::bailout ((!d_HF), "d_HF==NULL", "HeightFieldOps::rotate90");
	SanityCheck::bailout ((!d_HF->d_hf), "d_HF->d_hf==NULL", "HeightFieldOps::rotate90");

	PTYPE	*hft=NULL;
	int	w = d_HF->getWidth(), 
		h = d_HF->getHeight(),
		s = d_HF->getSize();
	char	buf[80];

	sprintf (buf, "Rotating %d degrees ...\n", degrees); 
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	if (degrees < 0)
		degrees = (-degrees)%360;
	else
	if (degrees > 360)
		degrees = degrees%360;

	if (degrees == 90 || degrees == 270)
		hft=new PTYPE [s];

	if (degrees == 0)
		{
		if (d_HF->d_hfBck)
			d_HF->restore ();
		}
	else
	if (degrees == 90)
		{
		for (int x=0; x<w; x++)
			for (int y=0; y<h; y++)
				hft[d_HF->Pos2Off(x,y)] = d_HF->El(h-1-y, x);
		}
	else
	if (degrees == 180)
		{
		PTYPE	pt;
		int	idx=s-1, 
			lim=s/2, 
			c=0;

		while (c < lim)
			{
			pt=d_HF->El(c);
			d_HF->d_hf[c]=d_HF->El(idx-c);
			d_HF->d_hf[idx-c]=pt;
			c++;
			}
		}
	else
	if (degrees == 270)
		{
		for (int x=0; x<w; x++)
			for (int y=0; y<h; y++)
				hft[d_HF->Pos2Off(x,y)] = d_HF->El(y, h-1-x);
		}
	else
		{
		printf ("Warning: Rotate90() paramter (%d) out of range ... Cancelling \n", degrees);
		return (-1);
		}

	if (degrees == 90 || degrees == 270)
		{
		memcpy (d_HF->d_hf, hft, sizeof(PTYPE)*d_HF->d_size);
		delete hft;
		}

	d_HF->setSaved (FALSE);
	return (0);
}


/*
 *  Invert: invert the height field 
 */
int HeightFieldOps::invert (int adjust_sea_level)
{
	char 	buf[80];
	SanityCheck::bailout ((!d_HF->d_hf), "d_HF->d_hf==NULL", "HeightFieldOps::invert");

	sprintf (buf, "Inverting height field ... \n");
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	for (int i=0; i<d_HF->d_size; i++)
		d_HF->d_hf[i]=1-d_HF->d_hf[i];
	d_HF->gatherStatistics ();
	if (adjust_sea_level)
		d_HF->setSealevel (1-d_HF->d_sealevel);
	d_HF->setSaved (FALSE);
	return (0);
}


/*
 *  hf_sort: compare 2 PTYPE values; used by qsort()
 */
inline int HeightFieldOps::hf_sort (const void *_v1_, const void *_v2_)
{
	PTYPE	*v1 = static_cast<PTYPE*>(_v1_),
		*v2 = static_cast<PTYPE*>(_v2_);

	if (*v1 < *v2)
		return (-1);
	else
	if (*v1 > *v2)
		return (1);
	else
		return (0);
}


/*
 *  fold: fold height field edges under sealevel.
 *  How: Take a slice 'margin' long, fill it with the ABS 
 *  difference between it's height and the HF, sort it and 
 *  then subtract it from the HF elevation at that point 
 *  (starting with the largest values on the HF outside), 
 *  scaling the value subtracted in a linear fashion (0..1).
 */
int HeightFieldOps::fold (int margin)
{
	int	i=0, lim=0, off, offlim, 
		(*hf_cmp_func) (const void *, const void *) = hf_sort;
	float	pstep; 
	PTYPE	slice[margin]; 
	char 	buf[80];

	SanityCheck::bailout ((!d_HF->d_hf), "d_HF->d_hf==NULL", "HeightFieldOps::Fold");
	SanityCheck::bailout ((margin<0), "margin<0", "HeightFieldOps::Fold");
	SanityCheck::bailout ((margin>d_HF->d_xsize/2 || margin>d_HF->d_ysize/2), "margin>x/2 || margin>y/2", "HeightFieldOps::Fold");

	sprintf (buf, "Folding with margin %d ...\n", margin); 
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	if (!margin)
		return (1);

	// top and bottom 
	off = 0;
	offlim = d_HF->d_xsize;
	for (int x=off; x<offlim; x++)
	    {
		//for (lim=0; lim<margin && d_HF->d_hf[d_HF->Pos2Off(x,lim)]>d_HF->d_sealevel; lim++)
		for (lim=0; lim<margin; lim++)
			{
			slice[lim] = d_HF->El(x,lim) - d_HF->d_sealevel;
			if (slice[lim] < 0)
				slice[lim] = 0;
			}
		qsort ((void *)slice, (size_t)lim, (size_t)sizeof(PTYPE), hf_cmp_func);
		pstep = (float)1/(float)lim;
		for (i=0; i<lim; i++)
			{
			d_HF->d_hf[d_HF->Pos2Off(x,i)] -= (slice[lim-1-i]*pstep*(lim-i));
			if (d_HF->El(x,i) < 0)
				d_HF->d_hf[d_HF->Pos2Off(x,i)] = 0;
			}
				

		//for (lim=0; lim<margin && d_HF->El(x,d_HF->d_ysize-1-lim)>d_HF->d_sealevel; lim++)
		for (lim=0; lim<margin; lim++)
			{
			slice[lim] = d_HF->El(x,d_HF->d_ysize-1-lim) - d_HF->d_sealevel;
			if (slice[lim] < 0)
				slice[lim] = 0;
			}
		qsort ((void *)slice, (size_t)lim, (size_t)sizeof(PTYPE), hf_cmp_func);
		pstep = (float)1/(float)lim;
		for (i=0; i<lim; i++)
			{
			d_HF->d_hf[d_HF->Pos2Off(x,d_HF->d_ysize-1-i)] -= (slice[lim-1-i]*pstep*(lim-i));
			if (d_HF->El(x,d_HF->d_ysize-1-i) < 0)
				d_HF->d_hf[d_HF->Pos2Off(x,d_HF->d_ysize-1-i)] = 0;
			}
	    off++;
	    offlim = d_HF->d_xsize;
	    }

	// left and right sides 
	for (int y=0; y<d_HF->d_ysize; y++)
	    {
		//for (lim=0; lim<margin && d_HF->El(lim,y)>d_HF->d_sealevel; lim++)
		for (lim=0; lim<margin; lim++)
			{
			slice[lim] = d_HF->El(lim,y) - d_HF->d_sealevel;
			if (slice[lim] < 0)
				slice[lim] = 0;
			}
		qsort ((void *)slice, (size_t)lim, (size_t)sizeof(PTYPE), hf_cmp_func);
		pstep = (float)1/(float)lim;
		for (i=0; i<lim; i++)
			{
			d_HF->d_hf[d_HF->Pos2Off(i,y)] -= (slice[lim-1-i]*pstep*(lim-i));
			if (d_HF->El(i,y) < 0)
				d_HF->d_hf[d_HF->Pos2Off(i,y)] = 0;
			}

		//for (lim=0; lim<margin && d_HF->El(d_HF->d_xsize-1-lim,y)>d_HF->d_sealevel; lim++)
		for (lim=0; lim<margin; lim++)
			{
			slice[lim] = d_HF->El(d_HF->d_xsize-1-lim,y) - d_HF->d_sealevel;
			if (slice[lim] < 0)
				slice[lim] = 0;
			}
		qsort ((void *)slice, (size_t)lim, (size_t)sizeof(PTYPE), hf_cmp_func);
		pstep = (float)1/(float)lim;
		for (i=0; i<lim; i++)
			{
			d_HF->d_hf[d_HF->Pos2Off(d_HF->d_ysize-1-i,y)] -= (slice[lim-1-i]*pstep*(lim-i));
			if (d_HF->El(d_HF->d_ysize-1-i,y) < 0)
				d_HF->d_hf[d_HF->Pos2Off(d_HF->d_ysize-1-i,y)] = 0;
			}
	    }

	d_HF->gatherStatistics ();
	d_HF->setSaved (FALSE);
	return (0);
}


/*
 *  linearScale: pick a point (centerx, centery) and scale the points 
 *  where distance is mindist<=distance<=maxdist linarly from either 
 *  1..0 if scale_factor is < 1; otherwise was scale from 1..2. 
 */
int HeightFieldOps::linearScale (	float 	centerx, 
					float	centery, 
					float 	scale_factor, 
					float 	mindist,
					float	maxdist, 
					float 	smoothFactor,
					bool 	invertScale,
					bool	scaleFullDist)
{
	float	p,		// percentage change multiplication factor
		smoothFloor,
		smoothCeil;
	double  d, 		// distance
		md;		// max distance
	char	buf[256];
		

	SanityCheck::bailout ((!d_HF->d_hf), "d_HF->d_hf==NULL", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((centerx<0), "centerx < 0", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((centerx>1), "centerx > 1", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((centery<0), "centery < 0", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((centery>1), "centery > 1", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((scale_factor<0), "scale_factor < 0", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((mindist<0), "mindist < 0", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((mindist>1), "mindist > 1", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((maxdist<0), "maxdist < 0", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((maxdist>1), "maxdist > 1", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((smoothFactor<0), "smoothFactor < 0", "HeightFieldOps::LinearScale");
	SanityCheck::bailout ((smoothFactor>1), "smoothFactor > 1", "HeightFieldOps::LinearScale");

	if (SanityCheck::warning ((maxdist==mindist), "maxdist==mindist - linear scale cancelled", NULL))
		return 1;

	sprintf (buf, "Linear Scaling (%.2f, %.2f, %.2f, %.2f, %.2f, %s, %s) ...\n",
		centerx, centery, scale_factor, mindist, maxdist, 
		invertScale==TRUE ? "InvertScale" : "NoInvertScale",
		scaleFullDist?"ScaleFullDist":"NoScaleFullDist");
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	// convert to real (subscript) scale
	centerx *= d_HF->d_xsize;
	centery *= d_HF->d_ysize;
	mindist *= (centerx+centery)/2;
	maxdist *= (centerx+centery)/2;
	//smoothFloor = mindist*(1-smoothFactor);
	//smoothCeil = maxdist*(1+smoothFactor);
	smoothFloor = mindist*(1+smoothFactor);
	smoothCeil = maxdist*(1-smoothFactor);

	md=MathTrig::distance (0, 0, d_HF->d_xsize, d_HF->d_ysize);
	//md=MathTrig::distance (centerx, centery, 0, 0);
	//if ((d=MathTrig::distance(centerx,centery,d_HF->d_xsize,0)) > md)
	//	md=d;
	//if ((d=MathTrig::distance(centerx,centery,0,d_HF->d_ysize)) > md)
	//	md=d;
	//if ((d=MathTrig::distance(centerx,centery,d_HF->d_xsize,d_HF->d_ysize)) > md)
	//	md=d;

	if (!scaleFullDist)
		md = maxdist;
		//md = (maxdist>md ? md-mindist : maxdist-mindist);

	for (int x=0; x<d_HF->d_xsize; x++)
		for (int y=0; y<d_HF->d_ysize; y++)
			{
			d=MathTrig::distance (centerx, centery, x, y);

			if (scale_factor >= 1)
				p=(d/md)*scale_factor;	// 1 .. scale_factor
			else
				p=(d/md)*(1-scale_factor);  // 0 .. 1

			if (smoothFactor > 0 && d>mindist && d<smoothFloor)
				{
				PTYPE	pt=0, pp;

				pp = (d-mindist)/(smoothFloor-mindist);
				pt += d_HF->El(x,y)*pp;
				pt += d_HF->El(x,y)*p*(1-pp);
				d_HF->d_hf[d_HF->Pos2Off(x,y)] = pt;
				}
			else
			if (smoothFactor > 0 && d<maxdist && d>smoothCeil)
				{
				PTYPE	pt=0, pp;

				pp = (maxdist-d)/(maxdist-smoothCeil);
				pt += d_HF->El(x,y)*p*pp;
				pt += d_HF->El(x,y)*(1-pp);
				d_HF->d_hf[d_HF->Pos2Off(x,y)] = pt;
				}
			else
			if (d >= smoothFloor && d <= smoothCeil)
				d_HF->d_hf[d_HF->Pos2Off(x,y)] *= p;

			//printf ("%f, %f\n", p, d_HF->El(x,y));
			//fflush (stdout);
			}

	d_HF->gatherStatistics ();
	d_HF->setSaved (FALSE);
	return (0);
}


/*
 *  mirror: mirror the height field on one of it's axis.
 *  Type can be 1, 2, 3, 4 for x-axis, y-axis, top (left
 *  to bottom right) or bottom (left to top right)
 */
int HeightFieldOps::mirror (int type)
{
	int	mid, 
		x, y,
		w = d_HF->d_xsize, 
		h = d_HF->d_ysize;
	PTYPE	t;
	

	// Mirror around X axis
	if (type == 1)
		{
		GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, 
			"Mirroring around X axis ...\n");

		mid = h/2;
		for (y=0; y<mid; y++)
		   for (x=0; x<w; x++)
			{
			t = d_HF->El(x,y);
			d_HF->d_hf[d_HF->Pos2Off(x,y)] = d_HF->El(x,h-y-1);
			d_HF->d_hf[d_HF->Pos2Off(x,h-y-1)] = t;
			}
		}
	else
	// Mirror around Y axis
	if (type == 2)
		{
		GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, 
			"Mirroring around Y axis ...\n");

		mid = w/2;
		for (x=0; x<mid; x++)
		    for (y=0; y<h; y++)
			{
			t = d_HF->El(x,y);
			d_HF->d_hf[d_HF->Pos2Off(x,y)] = d_HF->El(w-x-1,y);
			d_HF->d_hf[d_HF->Pos2Off(w-x-1,y)] = t;
			}
		}
	else
	// Mirror diagonally: this code will only work on square height fields
	if (type == 3 || type == 4)
		{
		if (w != h)
			{
			printf ("HeightFieldOps::mirror(): HF is not square ... cancelling\n");
			return (-1);
			}

		if (type == 3)
		    {
		    GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, 
			"Mirroring around TL-BR axis ...\n");
		    for (y=0; y<h; y++)
				for (x=y; x<w; x++)
					{
					t = d_HF->El(x,y);
					d_HF->d_hf[d_HF->Pos2Off(x,y)] = d_HF->El(y,x);
					d_HF->d_hf[d_HF->Pos2Off(y,x)] = t;
					}
		    }
		else
		    {
		    GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, 
			"Mirroring around BL-TR axis ...\n");
		    for (y=0; y<h; y++)
				for (x=0; x<w-y; x++)
					{
					t = d_HF->El(x,y);
					d_HF->d_hf[d_HF->Pos2Off(x,y)] = d_HF->El(w-y-1,h-x-1);
					d_HF->d_hf[d_HF->Pos2Off(w-y-1,h-x-1)] = t;
					}
		    }
		}
	else
		SanityCheck::bailout ((1), "Uncaught mirror type operation requested", "HeightFieldOps::mirror");

	return (0);
}


/*
 *  fill: fill the Height Field up to elv
 */
int HeightFieldOps::fill (PTYPE	elv, float factor, bool keepMin)
{
	int	n=0;
	float	p;
	char	buf[80];

	if (!d_HF->d_hf)
		{
		printf ("Warning: hf==NULL in HeightFieldOps::Fill ... operation cancelled\n");
		return (-1);
		}
	if (elv < 0 || elv > 1)
		{
		cerr << "Error: elevation out of range (" << 
			elv << ") in HeightFieldOps::Fill() ...\n";
		return (-1);
		}

	sprintf (buf, "Filling to elevation %f, delta %f%% ... ", elv, (1-factor)*100);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	if (keepMin)
		{
		for (int i=0; i<d_HF->d_size; i++)
			if (d_HF->El(i)<elv && d_HF->El(i)>d_HF->d_min)
				{
				p = elv-d_HF->El(i);
				p -= (p*factor);
				d_HF->d_hf[i] = elv - p;
				n++;
				}
	    	}
	else
		{
		for (int i=0; i<d_HF->d_size; i++)
			if (d_HF->El(i)<elv)
				{
				p = elv-d_HF->El(i);
				p -= (p*factor);
				d_HF->d_hf[i] = elv - p;
				n++;
				}
	   	}

	sprintf (buf, "filled %d points\n", n);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	d_HF->gatherStatistics ();
	d_HF->setSaved (FALSE);
	return (0);
}

//////////////////////////////////////////////////////////////////////////////

/*
 *  contour lines: generate contour lines (also called Isograms) and export 
 *	them to postscript. 
 * 
 * 	The return value is a pointer to a TFGListFlexArr* of FlexArray items 
 * 	which each holds a list of pointers to COORD2 structures (defined 
 * 	in MathTrig.h). 
 */
TFGListFlexArr* HeightFieldOps::contourLines (int nElvLevels, float epsilon, 
					int squareSize)
{
	SanityCheck::bailout ((!nElvLevels), "nElvLevels==0", "HeightFieldOps::HeightFieldOps");
	SanityCheck::bailout ((epsilon<=0), "epsolin<=0", "HeightFieldOps::HeightFieldOps");
	SanityCheck::bailout ((epsilon>=1), "epsolin>=1", "HeightFieldOps::HeightFieldOps");

	int		x, y, n, off;
	PTYPE		levelElv=1.0/(PTYPE)nElvLevels,	// elevation per level
			halfLevelElv=levelElv/2.0,	// half an elevation level
			el;
	char		buf[80];
	COORD2		*coord2=NULL;		// 2D coordinate
	TFGListFlexArr	*listMain = NULL;	// list of flexarrarys
	TFFlexArrayCoord2 *faCLine = NULL;	// flexarray of 2d coordinates

	// init variables
	d_clBuffer = new bool[d_HF->d_size];
	d_clPoints = 0;
	d_clLastRWidth = 0;

	sprintf (buf, "Finding %d contour lines with epsilon %.2f ... ", nElvLevels, epsilon);
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	d_clSquareSize = squareSize;
	memset (d_clBuffer, sizeof(bool)*d_HF->d_size, 0);
	listMain = new TFGListFlexArr ();
	d_clTargetElv = -levelElv/2.0;		// cE + lE = center of band

	// start looking for points on a contour line
	for (n=0; n<nElvLevels; n++)
	    {
	    d_clTargetElv += levelElv;
	    d_clLevelLow = d_clTargetElv - halfLevelElv*epsilon;
	    d_clLevelHigh = d_clTargetElv + halfLevelElv*epsilon;
	    for (y=0; y<d_HF->d_ysize; y++)
	      for (x=0; x<d_HF->d_xsize; x++)
		    {
		    off = d_HF->Pos2Off(x,y);
		    if (!d_clBuffer[off])
			  {
			  el = d_HF->El (off);
			  if (el >= d_clLevelLow && el <= d_clLevelHigh)
				{
				coord2 = new COORD2;
				coord2->x = x;
				coord2->y = y;
				faCLine = new TFFlexArrayCoord2 (100, 2);
				faCLine->append (coord2);
				d_clBuffer[off] = TRUE;
				clFollow (faCLine, x, y);

				if (faCLine->getSize() >= 3)
					{ 
					faCLine->trim ();
					listMain->append (faCLine);
					d_clPoints += faCLine->getSize();
					}
				else
					delete faCLine;
				}
			  }
	    	}
	     }

	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "done\n");
	sprintf (buf, "Found %d contour lines, %d points (avg. length = %.2f)\n", 
		listMain->length(), d_clPoints, ((float)d_clPoints)/((float)listMain->length()));
	GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

	delete [] d_clBuffer;
	d_clBuffer = NULL;

	return listMain;
}


/*
 *  clFollow: follow the contour line (Isogram) as if sniffing a 
 * 	trail. Check the cell's neighbours for and cells that could be 
 * 	on the countour line and recurse down this path, updating the 
 * 	coordinate cache (which is used as a marker to indicate that 
 * 	a certain point has already been used in a contour line). 
 */
void HeightFieldOps::clFollow (TFFlexArrayCoord2 *faCLine, int x, int y)
{
	SanityCheck::bailout ((!d_clBuffer), "d_clBuffer==NULL", "HeightFieldOps::clFollow");
	SanityCheck::bailout ((!faCLine), "faCLine==NULL", "HeightFieldOps::clFollow");

	PTYPE	*bestFit=NULL;
	COORD2	*coord2=NULL;		// 2D coordinate

	bestFit = clCheckSquareR (x, y, 1, d_clSquareSize);

	// store coordinates and continue recursing
	if (bestFit)
		{
		int	d = bestFit - d_HF->d_hf;

		coord2 = new COORD2;
		coord2->x = d_HF->Off2Posx(d);
		coord2->y = d_HF->Off2Posy(d);
		d_clBuffer[d] = TRUE;
		faCLine->append (coord2);
		clFollow (faCLine, coord2->x, coord2->y);
		}
}


/*
 *  clCheckSqaureR: recursiveley check a square starting with the 8 
 * 	immediate cell neightbours and then adding one to the offset 
 * 	print them. Stops recursing when a suitable cell is found.
 */
PTYPE *HeightFieldOps::clCheckSquareR (int x, int y, int width, int maxWidth)
{
	SanityCheck::bailout ((width<1), "width<1", "HeightFieldOps::clCheckSqaureR");
	SanityCheck::bailout ((width>maxWidth), "width>maxWidth", "HeightFieldOps::clCheckSqaureR");

	int 	i, j,
		iLow = (x >= width ? x-width : 0),
		iHigh = (x < (d_HF->d_xsize-width) ? x+width : d_HF->d_xsize-1),
		jLow = (y >= width ? y-width : 0),
		jHigh = (y < (d_HF->d_ysize-width) ? y+width : d_HF->d_ysize-1);
	PTYPE	*bestFit=NULL;

	//printf ("C: %d, %d,    %d, %d,    %d, %d\n", x, y, iLow, iHigh, jLow, jHigh);
	//fflush (stdout);
	// only scan outer lines (if last width was curr. width-1)
	if (d_clLastRWidth > 0 && d_clLastRWidth == width-1)	
	    {
	    for (i=iLow; i<=iHigh; i++)
		{
		bestFit = clCheckBestFit (bestFit, 
				&(d_HF->d_hf[d_HF->Pos2Off(i,jLow)]),
				d_HF->Pos2Off(i,jLow));
		bestFit = clCheckBestFit (bestFit, 
				&(d_HF->d_hf[d_HF->Pos2Off(i,jHigh)]),
				d_HF->Pos2Off(i,jHigh));
		}

	    for (j=jLow+1; j<=jHigh-1; j++)
		{
		bestFit = clCheckBestFit (bestFit, 
				&(d_HF->d_hf[d_HF->Pos2Off(iLow,j)]),
				d_HF->Pos2Off(iLow,j));
		bestFit = clCheckBestFit (bestFit, 
				&(d_HF->d_hf[d_HF->Pos2Off(iHigh,j)]),
				d_HF->Pos2Off(iHigh,j));
		}
	    }
	// scan whole square 
	else
	    for (i=iLow; i<=iHigh; i++)
		for (j=jLow; j<=jHigh; j++)
		    if (i != x || j!= y)
			bestFit = clCheckBestFit (bestFit, 
					&(d_HF->d_hf[d_HF->Pos2Off(i,j)]),
					d_HF->Pos2Off(i,j));

	d_clLastRWidth = width;
	if (width < maxWidth && !bestFit)
		bestFit = clCheckSquareR (x, y, width+1, maxWidth);
	else
	if (width >= maxWidth)
		d_clLastRWidth = 0;

	return bestFit;
}



/*
 *  clCheckBestFit: check if a new cell is better than the best so far
 */
PTYPE *HeightFieldOps::clCheckBestFit (PTYPE *bestFit, PTYPE *cell, int off)
{
	float	dold = d_clTargetElv-d_clLevelLow,
		dnew;
	PTYPE	el = d_HF->El(off); 

	if (!bestFit)
		{
		if (!d_clBuffer[off] &&
		     el >= d_clLevelLow && el <= d_clLevelHigh)
			bestFit = cell;
		}
	else
		{
		if (!d_clBuffer[off] &&
	    	     el >= d_clLevelLow && el <= d_clLevelHigh)
			{
			dnew = ABS (el-d_clTargetElv);
			if (dnew < dold)
				bestFit = cell;
			}
		}

	return bestFit;
}


//////////////////////////////////////////////////////////////////////////////


/* Add craters to a heightfield array 
 * Copyright (C) 1995 by Heiko Eissfeldt
 * heiko@colossus.escape.de
 * modified JAN 1996 John Beale for use with HF-Lab  
 * modified JUN 1999 Robert Gasch for use with Terraform
 */


#define alpha   (35.25/180.0*M_PI)      /* sphere segment cutoff angle */
#define crater_depth 0.85  /* 0.9 */


static double a1 = crater_depth, b, d;


/* 
 * crProfile: method to determine the z-elevation dependent on normalized 
 * 	squared radial coordinate 
 */
double HeightFieldOps::crProfile (double nsq_rad)
{
	static 	double c = 50.0; // 100.0;	// controls the wall thickeness 
	double 	radialcoord;	
	radialcoord = sqrt(fabs(nsq_rad));

	if (radialcoord > b) 
     		// outer region gauss distribution 
		return d*exp(-c*(radialcoord-b)*(radialcoord-b));
	else 
		// inner region sphere segment 
		return a1 - sqrt(1.0 - nsq_rad);
}



#define pow4(x) ((x)*(x)*(x)*(x))
#define border 0.63
#define inner_scale 1
#define outer_scale (1.0-inner_scale*pow4(border))/pow4(1.0-border)

/* 
 * crDissolve: function to control the weight of the crater profile as opposed
 * to the underlying terrain elevation 
 */
double HeightFieldOps::crDissolve (double nsq_rad)
{
#if 0
	if (nsq_rad > border*border) 
		{
		double 	temp = (1-sqrt(nsq_rad));
		return 1.0-outer_scale*pow4(temp);
		} 
	else 
		return inner_scale*nsq_rad*nsq_rad;
#else
	if (nsq_rad > 0.6*0.6) 
		return 1.0-0.9*exp(-30.0*(sqrt(nsq_rad) - 0.6)*(sqrt(nsq_rad) - 0.6));
	else 
		return 0.1*exp(-25.0*(sqrt(nsq_rad) - 0.6)*(sqrt(nsq_rad) - 0.6));
# if 0
	if (nsq_rad > border*border) 
		{
		double 	temp = (1-sqrt(nsq_rad));
		return 1.0-outer_scale*pow4(temp);
		} 
	else 
		{
		return 1.0- 4.0*(sqrt(nsq_rad) - 0.17)*(sqrt(nsq_rad) - 0.17);
		return 0.04+1.7*(sqrt(nsq_rad) - 0.38)*(sqrt(nsq_rad) - 0.38);
		}
# endif
#endif
}



//#define craterCoverage 0.15
//#define craterCoverage 0.50
#define X_WRAP(x)  x=(((x)>=0)?(x)%d_HF->d_xsize:(((x)%d_HF->d_xsize)+d_HF->d_xsize)%d_HF->d_xsize)
#define Y_WRAP(y)  y=(((y)>=0)?(y)%d_HF->d_ysize:(((y)%d_HF->d_ysize)+d_HF->d_ysize)%d_HF->d_ysize)


/* 
 * distributeCraters:  doing some damage to the surface 
 */
int HeightFieldOps::distributeCraters(unsigned int how_many, 
        bool wrap, D ch_scale, D cr_scale, D craterCoverage, int seed,
	float cx, float cy)
{
    int 	i, j, k, xloc, yloc, cratersize;
    int 	sq_radius;
    double 	nsq_rad, weight;
    int 	lev_samples, samples;
    double 	level_with, level_pure;
    double 	craterscale;		// vertical crater scaling factor 
    double 	c,d2, shift;
    double 	b1,b2,b3;		// radius scaling params 
    int 	meshsize;		// mean of xsize and ysize ? 
    int		lim = d_HF->getSize();
    PTYPE	*hfOrig = new PTYPE[lim];
    MathRandom	*rnd, *rndSeed;
    char	buf[80];

    SanityCheck::bailout ((how_many>1 && (cx!=-1||cy!=-1)), "impossible parameter combination", "HeightFieldOps::HeightFieldOps");
    SanityCheck::bailout ((how_many==1 && (cx<0||cy<0)), "impossible parameter combination", "HeightFieldOps::HeightFieldOps");

    sprintf (buf, "Distributing %d craters (%f, %f)\n", how_many,
		ch_scale, cr_scale);
    GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, buf);

    rnd = new MathRandom (seed);
    rndSeed = new MathRandom (rnd->rnd());
    b1 = 10.0;
    b2 = 1.0/b1/b1/b1/b1; 
    b3 = 1.0/b1;
    meshsize = (int)(sqrt(d_HF->d_xsize * d_HF->d_ysize));
    b = sin(alpha);
    d = a1 - cos(alpha);

    // copy the terrain 
    memcpy (hfOrig, d_HF->d_hf, lim*sizeof(PTYPE)); 

    k = how_many;
    while (k > 0) 
	{
	int 		ii, jj;
	MathRandom	*rndLoop = new MathRandom (rndSeed->rnd());

	/* pick a cratersize according to a power law distribution */
	/* greatest first. So great craters never eliminate small ones */
	/* c = (double)(how_many-k+1)/how_many + b3; */
	/* craters appear in random order */

	c = rnd->rnd1() + b3;	/* c is in the range b3 ... b3+1 */
	d2 = b2/c/c/c/c;	/* d2 is in the range 0 ... 1 */

	//if (how_many == 1) d2 = 1.0;   /* single craters set to max. size */

	if (how_many == 1)
		{
		xloc = (int)(cx*d_HF->d_xsize);
		yloc = (int)(cy*d_HF->d_xsize);
		cratersize = (int)(cr_scale*((d_HF->d_xsize+d_HF->d_ysize)/2));
		}
	else
		{
		xloc = (int)(rnd->rnd1()*d_HF->d_xsize);  /* pick a random location for the crater */
		yloc = (int)(rnd->rnd1()*d_HF->d_ysize);
		cratersize = 3 + (int)(d2 * craterCoverage * meshsize * cr_scale);
		}

/* macro to determine the height dependent on crater size */
#define CRATER_SCALE (((ch_scale*pow((cratersize/(3+craterCoverage*meshsize)),0.9)) \
		       /256*pow(meshsize/256.0,0.1))/craterCoverage*80)

	craterscale = CRATER_SCALE;     /* vertical crater scaling factor */
	
	/* what is the mean height of this plot */
	samples = lev_samples = MAX((cratersize*cratersize)/5, 1);
	level_with = level_pure = 0.0;
	while (samples) 
	    {
	    i = (int)(rndLoop->rnd1()*(2*cratersize) - cratersize);
	    j = (int)(rndLoop->rnd1()*(2*cratersize) - cratersize);

	    if (i*i+j*j > cratersize*cratersize) continue;
	    ii = xloc + i; jj = yloc + j;

	    // handle wrapping details... 
	    if ((wrap) || ((ii >= 0) && (ii < d_HF->d_xsize) && (jj >= 0) && (jj < d_HF->d_ysize))) 
		{
		X_WRAP(ii);
		Y_WRAP(jj);
		level_with += d_HF->El(ii, jj);
		level_pure += hfOrig[d_HF->Pos2Off(ii, jj)];
		samples--;
	    	}

	    }
	level_with /= lev_samples;
	level_pure /= lev_samples;

	/* Now lets create the crater. */

	/* In order to do it efficiently, we calculate for one octant
	 * only and use eightfold symmetry, if possible.
	 * Main diagonals and axes have a four fold symmetry only.
	 * The center point has to be treated single.
	 */

#define SQUEEZE 1.3

/* this macro calculates the coordinates, does clipping and modifies
 * the elevation at this spot due to shift and weight.
 * hfOrig() contains the crater-free underground. d_HF() contains the result.
 * level_with: average altitude of cratered surface in region
 * level_pure: average altitude of uncratered surface
 */

#define SHIFT(x,y) { \
    ii = xloc + (x); jj = yloc + (y); \
    if (wrap || ((ii >= 0) && (ii < d_HF->d_xsize) && \
		 (jj >= 0) && (jj < d_HF->d_ysize))) {\
      X_WRAP(ii);  Y_WRAP(jj); \
      d_HF->d_hf[d_HF->Pos2Off(ii,jj)] = (shift  + (d_HF->El(ii,jj))*weight + \
      (level_with + (hfOrig[d_HF->Pos2Off(ii, jj)]-level_pure)/SQUEEZE) * (1.0-weight)); \
    } \
  }

/* macro to do four points at once. Points are rotated by 90 degrees. */
#define FOURFOLD(i,j)   SHIFT(i,j) SHIFT(-j,i) SHIFT(-i,-j) SHIFT(j,-i)

/* get eightfold symmetry by mirroring fourfold symmetry along the x==y axe */
#define EIGHTFOLD       FOURFOLD(i,j) FOURFOLD(j,i)


	/* The loop covers a triangle (except the center point)
	 * Eg cratersize is 3, coordinates are shown as i,j
	 *
	 *              3,3 j
	 *         2,2  3,2 |
	 *    1,1  2,1  3,1 v
	 * x  1,0  2,0  3,0
	 * ^          <-i
	 * |
	 * center point
	 *
	 * 2,1 , 3,2 and 3,1 have eightfold symmetry.
	 * 1,0 , 2,0 , 3,0 , 1,1 , 2,2 and 3,3 have fourfold symmetry.
	 */

	for (i = cratersize; i > 0; i--) {
	    for (j = i; j >= 0; j--) {

		/* check if outside */
		sq_radius = i*i+j*j;
		nsq_rad = (double)sq_radius/cratersize/cratersize;
		if (nsq_rad > 1) continue;

		/* inside the crater area */
		shift = craterscale*crProfile(nsq_rad);
		weight = crDissolve(nsq_rad);

		if (i==j || j==0) {
		    FOURFOLD(i,j)
		} else {
		    EIGHTFOLD
		}
	    }
	}
	/* the center point */
	shift =  craterscale*crProfile(0.0);
	weight = crDissolve(0.0);
	SHIFT(0,0)

	/* one crater added */
	k--;

	if ((how_many - k) % 1000 == 0)
    		GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, ".");
	delete rndLoop;
	}

    delete hfOrig;
    delete rndSeed;
    delete rnd;
    d_HF->gatherStatistics ();
    d_HF->setSaved (FALSE);
    GlobalTrace::trace (GlobalTrace::TRACE_VERBOSE, "done\n");
    return(0);
} 

//////////////////////////////////////////////////////////////////////////////
